;;; ------------------------------------------------------------------------
;;;	CopyText.lsp v1.3
;;;
;;;	Copyright 03.16.09
;;;	Alan J. Thompson (alanjt)
;;;	alanjt@gmail.com
;;;
;;;	Permission to use, copy, modify, and distribute this software
;;;	for any purpose and without fee is hereby granted, provided
;;;	that the above copyright notice appears in all copies and
;;;	that both that copyright notice and the limited warranty and
;;;	restricted rights notice below appear in all supporting
;;;	documentation.
;;;
;;;	The following program(s) are provided "as is" and with all faults.
;;;	Alan J. Thompson DOES NOT warrant that the operation of the program(s)
;;;	will be uninterrupted and/or error free.
;;;
;;;	User has option to copy contents of selected text object or typed in
;;;	content, to any number of selected text objects (Text, Mtext, Multileader,
;;;	Attribute Definition). If only one target object is selected,
;;;	user is prompted with option to swap contents of source & target.
;;;	Will select objects in xrefs & blocks.
;;;
;;;	Revision History:
;;;
;;;	v1.1	(06.08.09) 1. Rewrite to code after replacing/updating subroutines.
;;;
;;;	v1.2	(09.18.09) 1. Added function to, if applicable, strip formatting codes
;;;			      of textstring taken from MtextAttributes.
;;;			   2. Removed issue with selecting Multileaders with blocks.
;;;			   3. Added option to select multiple 'target' text objects.
;;;			   4. Added option to type in text value to copy to others.
;;;			   5. Added subroutines: AT:SS->List, AT:TabFilter, AT:EditTextBox
;;;
;;;	v1.21	(09.21.09) 1. Updated AT:EditTextBox subroutine.
;;;
;;;	v1.3	(11.25.09) 1. Added subroutine AT:TextString and changed string extraction
;;;			      coding to extract string from DXF code, ensuring that all
;;;			      symbols are extracted properly (bug with vla-get-textstring).
;;;
;;; ------------------------------------------------------------------------

(defun c:CT (/) (c:CopyText))
(defun c:CopyText (/ *error* AT:Undo AT:Entsel AT:SS->List AT:TabFilter AT:TextString #Count
                   #SSPickfirst _#Strip #SourceObj #SourceText #TargetEnt #Flag #TargetObj
                   #TargetText #Answer
                  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; SUBROUTINES ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; error handler
  (defun *error* (#Message)
    (AT:Undo "V" "E")
    (and #Message
         (not (wcmatch (strcase #Message) "*BREAK*,*CANCEL*,*QUIT*"))
         (princ (strcat "\nError: " #Message))
    ) ;_ and
  ) ;_ defun



;;; Undo Begin/End (Either VLA or Command)
;;; #CommandVLA - "V" for VLA OR "C" for Command
;;; #BeginEnd - "B" for Undo Begin OR "E" for Undo End
;;; Alan J. Thompson, 03.23.09
  (defun AT:Undo (#CommandVLA #BeginEnd / #Cmdecho)
    (and (member (strcase #CommandVLA) '("C" "V"))
         (member (strcase #BeginEnd) '("B" "E"))
         (cond
           ((eq "C" (strcase #CommandVLA))
            (setq #Cmdecho (getvar 'cmdecho))
            (setvar 'cmdecho 0)
            (cond
              ((eq "B" (strcase #BeginEnd)) (command "_.undo" "_begin"))
              ((eq "E" (strcase #BeginEnd)) (command "_.undo" "_end"))
            ) ;_ cond
            (setvar 'cmdecho #Cmdecho)
           )
           ((eq "V" (strcase #CommandVLA))
            (or *AcadDoc* (setq *AcadDoc* (vla-get-activedocument (vlax-get-acad-object))))
            (cond
              ((eq "B" (strcase #BeginEnd)) (not (vla-StartUndoMark *AcadDoc*)))
              ((eq "E" (strcase #BeginEnd)) (not (vla-EndUndoMark *AcadDoc*)))
            ) ;_ cond
           )
         ) ;_ cond
    ) ;_ and
  ) ;_ defun



;;; Entsel or NEntsel with options
;;; #Nested - Entsel or Nentsel (T for Nentsel, nil for Entsel)
;;; #Message - Selection message (if nil, "\nSelect object: " is used)
;;; #FilterList - DXF ssget style filtering (nil if not required)
;;;               "V" as first item in list to convert object to VLA-OBJECT (must be in list if no DXF filtering)
;;;               "L" as first item in list to ignore locked layers (must be in list if no DXF filtering)
;;; #Keywords - Keywords to match instead of object selection (nil if not required)
;;; Example: (AT:Entsel nil "\nSelect MText not on 0 layer [Settings]: " '("LV" (0 . "MTEXT")(8 . "~0")) "Settings")
;;; Example: (AT:Entsel T "\nSelect object [Settings]: " '("LV") "Settings")
;;; Alan J. Thompson, 04.16.09
;;; Updated: Alan J. Thompson, 06.04.09 (changed filter coding to work as ssget style dxf filtering)
;;; Updated: Alan J. Thompson, 09.07.09 (added option to ignore locked layers and convert object to VLA-OBJECT
;;; Updated: Alan J. Thompson, 09.18.09 (fixed 'missed pick' alert)
  (defun AT:Entsel (#Nested #Message #FilterList #Keywords / #Count #Message #Choice #Ent
                    #VLA&Locked #FilterList
                   )
    (vl-load-com)
    (setvar "errno" 0)
    (setq #Count 0)
    ;; fix message
    (or #Message (setq #Message "\nSelect object: "))
    ;; set entsel/nentsel
    (if #Nested
      (setq #Choice nentsel)
      (setq #Choice entsel)
    ) ;_ if
    ;; check if option to convert to vla-object or ignore locked layers in #FilterList variable
    (and (vl-consp #FilterList)
         (eq (type (car #FilterList)) 'STR)
         (setq #VLA&Locked (car #FilterList)
               #FilterList (cdr #FilterList)
         ) ;_ setq
    ) ;_ and
    ;; select object
    (while (and (not #Ent) (/= (getvar "errno") 52))
      ;; if keywords
      (and #Keywords (initget #Keywords))
      (cond
        ((setq #Ent (#Choice #Message))
         ;; if ignore locked layers
         (and
           #VLA&Locked
           (vl-consp #Ent)
           (wcmatch (strcase #VLA&Locked) "*L*")
           (not
             (zerop
               (cdr (assoc 70
                           (entget (tblobjname
                                     "layer"
                                     (cdr (assoc 8 (entget (car #Ent))))
                                   ) ;_ tblobjname
                           ) ;_ entget
                    ) ;_ assoc
               ) ;_ cdr
             ) ;_ zerop
           ) ;_ not
           (setq #Ent nil
                 #Flag T
           ) ;_ setq
         ) ;_ and
         ;; #FilterList check
         (if (and #FilterList (vl-consp #Ent))
           ;; process filtering from #FilterList
           (or
             (not
               (member
                 nil
                 (mapcar
                   '(lambda (x)
                      (wcmatch
                        (strcase
                          (vl-princ-to-string
                            (cdr (assoc (car x) (entget (car #Ent))))
                          ) ;_ vl-princ-to-string
                        ) ;_ strcase
                        (strcase (vl-princ-to-string (cdr x)))
                      ) ;_ wcmatch
                    ) ;_ lambda
                   #FilterList
                 ) ;_ mapcar
               ) ;_ member
             ) ;_ not
             (setq #Ent nil
                   #Flag T
             ) ;_ setq
           ) ;_ or
         ) ;_ if
        )
      ) ;_ cond
      (and (or (= (getvar "errno") 7) #Flag)
           (/= (getvar "errno") 52)
           (not #Ent)
           (setq #Count (1+ #Count))
           (prompt (strcat "\nNope, keep trying!  "
                           (itoa #Count)
                           " missed pick(s)."
                   ) ;_ strcat
           ) ;_ prompt
      ) ;_ and
    ) ;_ while
    (if (and (vl-consp #Ent)
             #VLA&Locked
             (wcmatch (strcase #VLA&Locked) "*V*")
        ) ;_ and
      (vlax-ename->vla-object (car #Ent))
      #Ent
    ) ;_ if
  ) ;_ defun



;;; Edit text box (Old Mtext editor)
;;; Returns typed in textstring
;;; Alan J. Thompson, 09.18.09
  (defun AT:EditTextBox (/ *error* #Cmdecho #Mtexted #Mtext #String)
    (setq *error*  (lambda (msg)
                     (and #Mtext (entdel #Mtext))
                     (and #Cmdecho (setvar "cmdecho" #Cmdecho))
                     (and #Mtexted (setvar "mtexted" #Mtexted))
                   ) ;_ lambda
          #Cmdecho (getvar "cmdecho")
          #Mtexted (getvar "mtexted")
    ) ;_ setq
    (setvar "cmdecho" 0)
    (vl-catch-all-apply 'setvar (list "mtexted" "OldEditor"))
    (setq #Mtext (entmakex (list
                             '(0 . "MTEXT")
                             '(100 . "AcDbEntity")
                             '(100 . "AcDbMText")
                             (cons 10 (trans (cadr (grread t 4 4)) 1 0))
                           ) ;_ list
                 ) ;_ entmakex
    ) ;_ setq
    (vl-cmdf "_.mtedit" #Mtext)
    (setq #String (vla-get-textstring (vlax-ename->vla-object #Mtext)))
    (*error* nil)
    (if (/= #String "")
      #String
    ) ;_ if
  ) ;_ defun




;;; Convert selection set to list of ename or vla objects
;;; #Selection - SSGET selection set
;;; #VLAList - T for vla objects, nil for ename
;;; Alan J. Thompson, 04.20.09
  (defun AT:SS->List (#Selection #VlaList / #List)
    (and #Selection
         (setq #List (vl-remove-if
                       'listp
                       (mapcar 'cadr (ssnamex #Selection))
                     ) ;_ vl-remove-if
         ) ;_ setq
         #VlaList
         (setq #List (mapcar 'vlax-ename->vla-object #List))
    ) ;_ and
    #List
  ) ;_ defun



;;; Tab filter for ssget selection filtering
;;; Must use (list instead of '( to work
;;; Alan J. Thompson, 06.05.09
  (defun AT:TabFilter (/)
    (if (eq 2 (getvar "cvport"))
      (cons 410 "Model")
      (cons 410 (getvar "ctab"))
    ) ;_ if
  ) ;_ defun



;;; Extract textstring (with symbols) from text object
;;; Works on: MText, Text, Attrib, Attdef, Multileader
;;; #Ent - Ename to extract textstring from
;;; Alan J. Thompson, 11.24.09
  (defun AT:TextString (#Ent / #Ent #Str)
    (setq #Ent (entget #Ent)
          #Str ""
    ) ;_ setq
    (cond
      ((member (cdr (assoc 0 #Ent)) '("MTEXT" "TEXT" "ATTRIB" "ATTDEF"))
       (cond
         ((assoc 3 #Ent)
          (foreach x #Ent
            (if (eq 3 (car x))
              (setq #Str (strcat #Str (cdr x)))
            ) ;_ if
          ) ;_ foreach
          (setq #Str (strcat #Str (cdr (assoc 1 #Ent))))
         )
         (T (setq #Str (cdr (assoc 1 #Ent))))
       ) ;_ cond
      )
      ((eq (cdr (assoc 0 #Ent)) "MULTILEADER")
       (setq #Str (cdr (assoc 304 #Ent)))
      )
    ) ;_ cond
  ) ;_ defun



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; MAIN ROUTINE ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


  (vl-load-com)

  (AT:Undo "V" "B")
  (setq #Count 0)
  (if (setq #SSPickfirst
             (ssget "_I" '((0 . "MTEXT,TEXT,MULTILEADER,ATTDEF")))
      ) ;_ setq
    (setq #TargetEnt "Multiple")
  ) ;_ if

  ;; format strip for Multiline Attributes, if applicable
  (setq _#Strip
         (lambda (#String / #String)
           (mapcar '(lambda (f r)
                      (while (vl-string-search f #String)
                        (setq #String (vl-string-subst r f #String))
                      ) ;_ while
                    ) ;_ lambda
                   ;; find list
                   (list "\\P" "\\L" "\\l" "\\O")
                   ;; replace list
                   (list " " "" "" "")
           ) ;_ mapcar
           (vl-string-right-trim
             "}"
             (vl-string-left-trim "{" #String)
           ) ;_ vl-string-right-trim
         ) ;_ lambda
  ) ;_ setq

  (if
    (and
      (setq #SourceObj
             (AT:Entsel
               T
               "\nSelect source text object for copying or [Type]: "
               '("LV"
                 (0 . "MTEXT,TEXT,MULTILEADER,ATTDEF,ATTRIB")
                )
               "Type"
             ) ;_ AT:Entsel
      ) ;_ setq
      (cond
        ;; text object selected
        ((eq (type #SourceObj) 'VLA-OBJECT)
         (not (vl-catch-all-error-p
                (setq #SourceText
                       (vl-catch-all-apply
                         'AT:TextString
                         (list (vlax-vla-object->ename #SourceObj))
                       ) ;_ vl-catch-all-apply
                ) ;_ setq
              ) ;_ vl-catch-all-error-p
         ) ;_ not
        )
        ;; "Type" option
        ((eq #SourceObj "Type") (setq #SourceText (AT:EditTextBox)))
      ) ;_ cond
    ) ;_ and
     (progn
       ;; time to select text object whose contents will be replaced with source
       (while (and (not #Flag)
                   (or #SSPickfirst
                       (setq #TargetEnt
                              (AT:Entsel
                                T
                                "\nSelect target text object to modify or [Multiple]: "
                                '("L"
                                  (0 . "MTEXT,TEXT,MULTILEADER,ATTDEF,ATTRIB")
                                 )
                                "Multiple"
                              ) ;_ AT:Entsel
                       ) ;_ setq
                   ) ;_ or
              ) ;_ and
         (if (or
               ;; multiple selection
               (and (eq #TargetEnt "Multiple")
                    (setq #TargetObj
                           (AT:SS->List
                             (ssget
                               ":L"
                               (list
                                 (AT:TabFilter)
                                 (cons 0 "MTEXT,TEXT,MULTILEADER,ATTDEF")
                               ) ;_ list
                             ) ;_ ssget
                             T
                           ) ;_ AT:SS->List
                    ) ;_ setq
                    (setq #Flag T)
               ) ;_ and
               ;; single selection
               (and (vl-consp #TargetEnt)
                    (setq #TargetObj
                           (list (vlax-ename->vla-object (car #TargetEnt))
                           ) ;_ list
                    ) ;_ setq
               ) ;_ and
             ) ;_ or
           (foreach x #TargetObj
             (cond
               ;; source & target are same object
               ((and (eq (type #SourceObj) 'VLA-OBJECT)
                     (eq (vlax-get-property #SourceObj 'ObjectID)
                         (vlax-get-property x 'ObjectID)
                     ) ;_ eq
                ) ;_ and
                (and (vl-consp #TargetEnt)
                     (alert "Cannot match source with itself!")
                ) ;_ and
               )
               ;; able to extract textstring
               ((not (vl-catch-all-error-p
                       (setq #TargetText
                              (vl-catch-all-apply
                                'AT:TextString
                                (list (vlax-vla-object->ename x))
                              ) ;_ vl-catch-all-apply
                       ) ;_ setq
                     ) ;_ vl-catch-all-error-p
                ) ;_ not
                (setq #Count (1+ #Count))
                (vl-catch-all-apply
                  'vlax-put-property
                  (list
                    x
                    'TextString
                    ;; account for formattting, if going from multiline attribute to text or single line attribute
                    (if
                      (and
                        ;; source is MtextAttribute
                        (eq (vl-catch-all-apply
                              'vlax-get-property
                              (list #SourceObj 'MTextAttribute)
                            ) ;_ vl-catch-all-apply
                            :vlax-true
                        ) ;_ eq
                        ;; target not MtextAttribute
                        (not (eq (vl-catch-all-apply
                                   'vlax-get-property
                                   (list x 'MTextAttribute)
                                 ) ;_ vl-catch-all-apply
                                 :vlax-true
                             ) ;_ eq
                        ) ;_ not
                        ;; target not Mtext or multileader
                        (not (member (vlax-get-property x 'ObjectName)
                                     (list "AcDbMText" "AcDbMLeader")
                             ) ;_ member
                        ) ;_ not
                      ) ;_ and
                       (_#Strip #SourceText)
                       #SourceText
                    ) ;_ if
                  ) ;_ list
                ) ;_ vl-catch-all-apply
               )
               ;; unable to extract text, must not be able to edit it
               (T
                (and (vl-consp #TargetEnt)
                     (alert "Cannot edit text within selected object!")
                ) ;_ and
               )
             ) ;_ cond
           ) ;_ foreach
         ) ;_ if
       ) ;_ while
       ;; if only 1 target object selected, option to switch source & target offered
       (and
         (eq 1 #Count)
         (not (eq (type #SourceObj) 'STR))
         (not (initget 0 "Yes No Swap"))
         (setq
           #Answer (getkword
                     "\nSwap source & target contents? [Yes/No] <No>: "
                   ) ;_ getkword
         ) ;_ setq
         (member #Answer (list "Yes" "Swap"))
         (vl-catch-all-apply
           'vlax-put-property
           (list
             #SourceObj
             'TextString
             ;; account for formatting, if going from multiline attribute to text or single line attribute
             (if
               (and
                 (eq 1 (length #TargetObj))
                 ;; target is MtextAttribute
                 (eq (vl-catch-all-apply
                       'vlax-get-property
                       (list (car #TargetObj) 'MTextAttribute)
                     ) ;_ vl-catch-all-apply
                     :vlax-true
                 ) ;_ eq
                 ;; source not MtextAttribute
                 (not (eq (vl-catch-all-apply
                            'vlax-get-property
                            (list #SourceObj 'MTextAttribute)
                          ) ;_ vl-catch-all-apply
                          :vlax-true
                      ) ;_ eq
                 ) ;_ not
                 ;; source not Mtext or multileader
                 (not (member (vlax-get-property #SourceObj 'ObjectName)
                              (list "AcDbMText" "AcDbMLeader")
                      ) ;_ member
                 ) ;_ not
               ) ;_ and
                (_#Strip #TargetText)
                #TargetText
             ) ;_ if
           ) ;_ list
         ) ;_ vl-catch-all-apply
       ) ;_ and
     ) ;_ progn
  ) ;_ if

  (*error* nil)

  (princ)
) ;_ defun